#!/usr/bin/env python

import argparse
import os
import subprocess
import sys
import timeit

def RunCommand(command, print_stdout=False):
  options = {}

  # If using posix make sure started process is in a different
  # process group so that if the started process does something
  # crazy like... killing all processes in its process group 
  # we don't end up being killed ourself.
  if os.name == 'posix':
    options['preexec_fn'] = os.setsid

  processInstance=subprocess.Popen(command,
    stdout=subprocess.PIPE, 
    stderr=subprocess.STDOUT,
    **options
  )
  stdout=processInstance.communicate() #Allow program to run and wait for it to exit.    
  if print_stdout:
    print stdout[0]
  return processInstance.returncode == 0

def PrintResults(tester, kernel, power, runs, succeeded):
  row = [ tester, kernel, '2**%d' % power ]
  row.extend([ str(x) for x in runs ])
  if succeeded:
    avg = sum(runs)/len(runs)
    row.append(str(avg))
  else:
    row.append('Fail')
  print ', '.join(row)

def RunExperiment(tester, kernel, maxpower, numruns=3, extraflags=''):
  testerCommand = [ tester ]

  for power in range(1,maxpower+1):
    nelements = 2**power
   
    runs = []
    succeeded = True
    for runNumber in range(numruns):

      start = timeit.default_timer()
      succeeded &= RunCommand(testerCommand + [ str(nelements), kernel ] + extraflags.split(' '))
      if not succeeded:
        print 'Tester failed'
        break
      end = timeit.default_timer()
      runs.append(end-start)

    PrintResults(tester, kernel, power, runs, succeeded)

def main():
  parser = argparse.ArgumentParser(description='Run experiments')
  parser.add_argument('--numruns', metavar='N', type=int, nargs=1, help='number of runs per test', default=3)
  parser.add_argument('--atest-flags', metavar='str', type=str, help='extra flags to pass to atester', default='')
  parser.add_argument('--vtest-flags', metavar='str', type=str, help='extra flags to pass to vtester', default='')
  args = parser.parse_args()

  print "Running experiments"
  print "   numruns    : %d" % args.numruns
  print "   atest-flags: %s" % args.atest_flags
  print "   vtest-flags: %s" % args.vtest_flags

  os.chdir('src')
  RunCommand(['./atester', '4', 'blelloch.cl', '--clinfo'], print_stdout=True)
  for (t,maxpower,extra) in [ ('./atester',20,args.atest_flags), ('./vtester',10,args.vtest_flags) ]:
    for k in ["brentkung.cl", "blelloch.cl", "koggestone.cl", "sklansky.cl"]:
      RunExperiment(t, k, maxpower, args.numruns, extra)

  return 0

if __name__ == '__main__':
  sys.exit(main())
