#define ABSTRACT
#include "clwrapper.h"
#include "abstraction.h"

#include <cassert>
#include <cstdio>
#include <iomanip>
#include <iostream>
#include <map>
#include <sstream>
#include <string>
#include <time.h>

#define ispow2(x) ((x & (x-2)) == 0)
#define MAX_GROUP_SIZE 1024

using namespace std;

void print_array(TYPE *xs, const char *name, unsigned len) {
  for (unsigned i=0; i<len; ++i) {
    cout << name << "[" << i << "] = (" 
              << GET_LOWER(xs[i]) << ", " << GET_UPPER(xs[i]) << ")"
              << " [ " << hex << xs[i] << dec << "]"
              << endl;
  }
}

void print_usage(char *progname) {
  cerr << "Usage: " << progname << " <nelements> <kernel.cl> [options]" << endl;
  cerr << "  where <nelements> must be a power of two" << endl;
  cerr << "        <kernel.cl> must be one of" << endl;
  cerr << "          o blelloch.cl" << endl;
  cerr << "          o brentkung.cl" << endl;
  cerr << "          o koggestone.cl" << endl;
  cerr << "          o sklansky.cl" << endl;
  cerr << "  --clinfo           print platform/device info" << endl;
  cerr << "  --is-exclusive     invoke scan as exclusive prefix sum" << endl;
  cerr << "  --check-random     force random check at end of kernel" << endl;
  cerr << "  --print-results    print results" << endl;
  cerr << "  --platform=X       use platform X (default 0)" << endl;
  cerr << "  --device=Y         use device Y (default 0)" << endl;
}

int main(int argc, char **argv) {
  // command line arguments
  if (argc < 3) {
    print_usage(argv[0]);
    return 1;
  }

  // total number of elements
  unsigned nelements = atoi(argv[1]);
  if (!ispow2(nelements)) {
    cerr << "Error: nelements must be a power of two" << endl;
    return 1;
  }

  // number of elements per group
  unsigned N;
  // number of groups
  unsigned ngroups;
  if (nelements <= MAX_GROUP_SIZE) {
    N = nelements;
    ngroups = 1;
  } else {
    N = MAX_GROUP_SIZE;
    ngroups = nelements/N;
  }
  cout << "nelements = " << nelements << endl;
  cout << "N         = " << N << endl;
  cout << "ngroups   = " << ngroups << endl;

  bool print_clinfo = false;
  bool print_results = false;
  bool check_random = false;
  unsigned is_exclusive = 0;
  bool force_host_scan = false;
  unsigned platform = 0;
  unsigned device = 0;
  for (int i=3; i<argc; ++i) {
    string opt = string(argv[i]);
    string platform_prefix = "--platform=";
    string device_prefix   = "--device=";
    if (opt == "-h" || opt == "--help") {
      print_usage(argv[0]);
      return 0;
    } else if (opt == "--check-random") {
      check_random = true;
    } else if (opt == "--clinfo") {
      print_clinfo = true;
    } else if (opt == "--is-exclusive" || opt == "--is-excl") {
      is_exclusive = 1;
    } else if (opt == "--force-host-scan") {
      force_host_scan = true;
    } else if (opt == "--print-results") {
      print_results = true;
    } else if ((!opt.compare(0, platform_prefix.size(), platform_prefix))) {
      platform = (unsigned) atoi(opt.substr(platform_prefix.size()).c_str());
    } else if ((!opt.compare(0, device_prefix.size(), device_prefix))) {
      device = (unsigned) atoi(opt.substr(device_prefix.size()).c_str());
    } else {
      cout << "Unrecognised option: " << opt << endl;
    }
  }

  // kernel specific parameters
  string kernel = string(argv[2]);
  if (kernel != "sklansky.cl" &&
      kernel != "koggestone.cl" &&
      kernel != "brentkung.cl" &&
      kernel != "blelloch.cl") {
    cerr << "Error: kernel not unrecognised " << kernel << endl;
    return 1;
  }

  // platform info
  if (print_clinfo) {
    cout << clinfo();
  }

  // initialise for device on platform, with profiling off
  // this creates a context and command queue
  bool profiling = false;
  CLWrapper clw(platform, device, profiling);

  // compile the OpenCL code
  cl_program meta;
  {
    ostringstream oss;
    oss << "-I. -DABSTRACT -DNO_INVARIANTS -DN=" << N << " -DINNER=" << kernel << (is_exclusive ? " -DEXCLUSIVE" : "");
    meta = clw.compile("meta.cl", oss.str().c_str());
  }

  cl_program scan;
  if (1 < ngroups && ngroups <= MAX_GROUP_SIZE) {
    ostringstream oss;
    oss << "-I. -DABSTRACT -DNO_INVARIANTS -DN=" << ngroups;
    scan = clw.compile(kernel.c_str(), oss.str().c_str());
  }

  // create some memory objects on the device
  size_t ArraySize = nelements * sizeof(TYPE);
  size_t SumSize   = ngroups   * sizeof(TYPE);
  cout << "sizeof(TYPE) = " << sizeof(TYPE) << endl;
  cout << "ArraySize    = " << ArraySize    << endl;
  cl_mem d_in     = clw.dev_malloc(ArraySize);
  cl_mem d_out    = clw.dev_malloc(ArraySize);
  cl_mem d_sum    = clw.dev_malloc(SumSize);
  cl_mem d_sumout = clw.dev_malloc(SumSize);
  cl_mem d_error  = clw.dev_malloc(sizeof(unsigned));

  // initialise input
  {
    cl_kernel k = clw.create_kernel(meta, "init_abstract");
    clw.kernel_arg(k, d_in);
    cl_uint dim = 1;
    size_t global_work_size = nelements;
    size_t local_work_size = N;
    clw.run_kernel(k, dim, &global_work_size, &local_work_size);
  }

  if (print_results) {
    TYPE *in = (TYPE *)malloc(ArraySize);
    assert(in);
    clw.memcpy_from_dev(d_in, ArraySize, in);
    print_array(in, "in", nelements);
    free(in);
  }

  // block level scan
  {
  size_t global_work_size;
  size_t local_work_size;
  if (kernel == "sklansky.cl") {
    global_work_size = nelements/2;
    local_work_size  = N/2;
  } else if (kernel == "koggestone.cl") {
    global_work_size = nelements;
    local_work_size  = N;
  } else if (kernel == "brentkung.cl") {
    global_work_size = nelements/2;
    local_work_size  = N/2;
  } else if (kernel == "blelloch.cl") {
    global_work_size = nelements/2;
    local_work_size  = N/2;
  } else {
    assert(0);
  }
  cl_kernel k = clw.create_kernel(meta, "meta1");
  clw.kernel_arg(k, d_in, d_out, d_sum, is_exclusive);
  cl_uint dim = 1;
  clw.run_kernel(k, dim, &global_work_size, &local_work_size);
  }

  if (1 < ngroups) {
    if (ngroups <= MAX_GROUP_SIZE && !force_host_scan) {
      // device level exclusive scan
      cl_uint dim = 1;
      size_t global_work_size;
      size_t local_work_size;
      if (kernel == "sklansky.cl") {
        global_work_size = ngroups/2;
        local_work_size  = ngroups/2;
      } else if (kernel == "koggestone.cl") {
        global_work_size = ngroups;
        local_work_size  = ngroups;
      } else if (kernel == "brentkung.cl") {
        global_work_size = ngroups/2;
        local_work_size  = ngroups/2;
      } else if (kernel == "blelloch.cl") {
        global_work_size = ngroups/2;
        local_work_size  = ngroups/2;
      } else {
        assert(0);
      }
      unsigned is_exclusive = 1;
      cl_kernel k = clw.create_kernel(scan, "prefixsum");
      clw.kernel_arg(k, d_sum, d_sumout, is_exclusive);
      clw.run_kernel(k, dim, &global_work_size, &local_work_size);
    } else {
      // host level exclusive scan
      TYPE *sum = (TYPE *)malloc(SumSize);
      assert(sum);
      TYPE *res = (TYPE *)malloc(SumSize);
      assert(res);
      clw.memcpy_from_dev(d_sum, SumSize, sum);
      res[0] = IDENTITY;
      for (unsigned i=1; i<ngroups; ++i) {
        res[i] = OPERATOR(res[i-1], sum[i-1]);
      }
      if (print_results) {
        TYPE *out = (TYPE *)malloc(ArraySize);
        clw.memcpy_from_dev(d_out, ArraySize, out);
        print_array(out, "out", nelements);
        free(out);
        print_array(sum, "sum", ngroups);
        print_array(res, "res", ngroups);
      }
      clw.memcpy_to_dev(d_sumout, SumSize, res);
      free(sum);
      free(res);
    }

    // final increment
    {
      size_t global_work_size;
      size_t local_work_size;
      global_work_size = nelements;
      local_work_size  = N;
      cl_kernel k = clw.create_kernel(meta, "meta2");
      clw.kernel_arg(k, d_out, d_sumout);
      cl_uint dim = 1;
      clw.run_kernel(k, dim, &global_work_size, &local_work_size);
    }
  }

  if (print_results) {
    TYPE *out = (TYPE *)malloc(ArraySize);
    clw.memcpy_from_dev(d_out, ArraySize, out);
    print_array(out, "out", nelements);
    free(out);
  }

  // check results
  unsigned error = 0;
  {
    clw.memcpy_to_dev(d_error, sizeof(unsigned), &error);
    cl_kernel k = clw.create_kernel(meta, is_exclusive ?  "check_abstract_exclusive" : "check_abstract_inclusive");
    clw.kernel_arg(k, d_out, d_error);
    cl_uint dim = 1;
    size_t global_work_size = nelements;
    size_t local_work_size = N;
    clw.run_kernel(k, dim, &global_work_size, &local_work_size);
    clw.memcpy_from_dev(d_error, sizeof(unsigned), &error);
    if (error == 0) {
      printf("TEST PASSED (%s)\n", is_exclusive ? "EXCL" : "INCL");
    } else {
      printf("TEST FAILED\n");
    }
  }

  // random check
  if (check_random) {
    srand(time(NULL));
    TYPE *out = (TYPE *)malloc(ArraySize);
    clw.memcpy_from_dev(d_out, ArraySize, out);
    unsigned i = (rand() % (int)(nelements + 1));
    cout << "Random check" << endl;
    cout << "out" << "[" << i << "] = (" << GET_LOWER(out[i]) << ", " << GET_UPPER(out[i]) << ")" << " [ " << hex << out[i] << dec << "]" << endl;
    assert(out[i] == (is_exclusive && i == 0) ? IDENTITY : MAKE_PAIR(0, is_exclusive ? i : i+1));
    free(out);
  }

  return error;
}

