#include "clwrapper.h"

#include <cassert>
#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <time.h>

#define ispow2(x) ((x & (x-2)) == 0)
#define MAX_GROUP_SIZE 1024
#define TYPE char

using namespace std;

// the nth triangular number (the number of op1 vectors)
unsigned triangular(unsigned n) {
  return (n * (n+1))/2;
}

void print_vector(TYPE *v, TYPE *e, unsigned nelements) {
  printf("v = {%d", v[0]);
  for (unsigned i=1; i<nelements; ++i) {
    printf(", %d", v[i]);
  }
  printf("}\t");
  printf("e = {%d", e[0]);
  for (unsigned i=1; i<nelements; ++i) {
    printf(", %d", e[i]);
  }
  printf("}\n");
}

unsigned runvector(CLWrapper &clw,
                   string kernel,
                   cl_kernel &k, cl_kernel &kscan, cl_kernel &kinc, cl_kernel &kcheck,
                   TYPE *v, TYPE *e,
                   cl_mem &d_in, cl_mem &d_out, cl_mem &d_sum, cl_mem &d_sumout, cl_mem &d_expected, cl_mem &d_error,
                   unsigned nelements, unsigned N, unsigned ngroups,
                   bool force_host_scan, bool print_vectors, bool print_results, bool check_random) {
  size_t ArraySize = nelements * sizeof(TYPE);

  if (print_vectors) {
    print_vector(v, e, nelements);
  }

  // block level scan
  {
    cl_uint dim = 1;
    size_t global_work_size;
    size_t local_work_size;
    if (kernel == "sklansky.cl") {
      global_work_size = nelements/2;
      local_work_size  = N/2;
    } else if (kernel == "koggestone.cl") {
      global_work_size = nelements;
      local_work_size  = N;
    } else if (kernel == "brentkung.cl") {
      global_work_size = nelements/2;
      local_work_size  = N/2;
    } else if (kernel == "blelloch.cl") {
      global_work_size = nelements/2;
      local_work_size  = N/2;
    } else {
      assert(0);
    }
    clw.memcpy_to_dev(d_in, ArraySize, v);
    clw.kernel_arg(k, d_in, d_out, d_sum);
    clw.run_kernel(k, dim, &global_work_size, &local_work_size);
  }

  // nb: if ngroups > MAX_GROUP_SIZE then the test will fail
  if (1 < ngroups) {
    if (ngroups <= MAX_GROUP_SIZE && !force_host_scan) {
      // device level exclusive scan
      cl_uint dim = 1;
      size_t global_work_size;
      size_t local_work_size;
      if (kernel == "sklansky.cl") {
        global_work_size = ngroups/2;
        local_work_size  = ngroups/2;
      } else if (kernel == "koggestone.cl") {
        global_work_size = ngroups;
        local_work_size  = ngroups;
      } else if (kernel == "brentkung.cl") {
        global_work_size = ngroups/2;
        local_work_size  = ngroups/2;
      } else if (kernel == "blelloch.cl") {
        global_work_size = ngroups/2;
        local_work_size  = ngroups/2;
      } else {
        assert(0);
      }
      unsigned is_exclusive = 1;
      clw.kernel_arg(kscan, d_sum, d_sumout, is_exclusive);
      clw.run_kernel(kscan, dim, &global_work_size, &local_work_size);
    } else {
      // host level exclusive scan
      assert(0);
    }

    // final increment
    {
      cl_uint dim = 1;
      size_t global_work_size = nelements;
      size_t local_work_size = N;
      clw.kernel_arg(kinc, d_out, d_sumout);
      clw.run_kernel(kinc, dim, &global_work_size, &local_work_size);
    }
  }

  // check results
  unsigned error = 0;
  {
    cl_uint dim = 1;
    size_t global_work_size = nelements;
    size_t local_work_size  = N;
    clw.memcpy_to_dev(d_expected, ArraySize, e);
    clw.memcpy_to_dev(d_error, sizeof(unsigned), &error);
    clw.kernel_arg(kcheck, d_out, d_expected, d_error);
    clw.run_kernel(kcheck, dim, &global_work_size, &local_work_size);
    clw.memcpy_from_dev(d_error, sizeof(unsigned), &error);
  }

  if (print_results) {
    TYPE *out = (TYPE *)malloc(ArraySize); assert(out);
    clw.memcpy_from_dev(d_out, ArraySize, out);
    printf("out = {%d", out[0]);
    for (unsigned i=1; i<nelements; ++i) {
      printf(", %d", out[i]);
    }
    printf("}\n");
    free(out);
  }

  if (check_random) {
    srand(time(NULL));
    TYPE *out = (TYPE *)malloc(ArraySize);
    clw.memcpy_from_dev(d_out, ArraySize, out);
    unsigned i = (rand() % (int)(nelements + 1));
    printf("random-check: out[%d] = %d\n", i, out[i]);
    assert(out[i] == e[i]);
    free(out);
  }

  return error;
}

void print_usage(char *progname) {
  cerr << "Usage: " << progname << " <nelements> <kernel.cl> [options]" << endl;
  cerr << "  where <nelements> must be a power of two" << endl;
  cerr << "        <kernel.cl> must be one of" << endl;
  cerr << "          o blelloch.cl" << endl;
  cerr << "          o brentkung.cl" << endl;
  cerr << "          o koggestone.cl" << endl;
  cerr << "          o sklansky.cl" << endl;
  cerr << "  --clinfo           print platform/device info" << endl;
  cerr << "  --check-random     force random check at end of each kernel invocation" << endl;
  cerr << "  --print-vectors    print generated test vectors" << endl;
  cerr << "  --print-results    print results" << endl;
  cerr << "  --platform=X       use platform X (default 0)" << endl;
  cerr << "  --device=Y         use device Y (default 0)" << endl;
}

int main(int argc, char **argv) {
  if (argc < 3) {
    print_usage(argv[0]);
    return 1;
  }

  // total number of elements (per testvector)
  unsigned nelements = atoi(argv[1]);
  if (!ispow2(nelements)) {
    cerr << "Error: nelements must be a power of two" << endl;
    return 1;
  }

  string kernel = string(argv[2]);
  if (kernel != "sklansky.cl" &&
      kernel != "koggestone.cl" &&
      kernel != "brentkung.cl" &&
      kernel != "blelloch.cl") {
    cerr << "Error: kernel not unrecognised " << kernel << endl;
    return 1;
  }

  bool print_clinfo = false;
  bool print_vectors = false;
  bool print_results = false;
  bool force_host_scan = false;
  bool check_random = false;
  unsigned platform = 0;
  unsigned device = 0;
  for (int i=3; i<argc; ++i) {
    string opt = string(argv[i]);
    string platform_prefix = "--platform=";
    string device_prefix   = "--device=";
    if (opt == "-h" || opt == "--help") {
      print_usage(argv[0]);
      return 0;
    } else if (opt == "--clinfo") {
      print_clinfo = true;
    } else if (opt == "--print-vectors") {
      print_vectors = true;
    } else if (opt == "--print-results") {
      print_results = true;
    } else if (opt == "--check-random") {
      check_random = true;
    } else if (opt == "--force-host-scan") {
      force_host_scan = true;
    } else if ((!opt.compare(0, platform_prefix.size(), platform_prefix))) {
      platform = (unsigned) atoi(opt.substr(platform_prefix.size()).c_str());
    } else if ((!opt.compare(0, device_prefix.size(), device_prefix))) {
      device = (unsigned) atoi(opt.substr(device_prefix.size()).c_str());
    } else {
      cout << "Unrecognised option: " << opt << endl;
    }
  }

  // number of elements per group
  unsigned N;
  // number of groups
  unsigned ngroups;
  if (nelements <= MAX_GROUP_SIZE) {
    N = nelements;
    ngroups = 1;
  } else {
    N = MAX_GROUP_SIZE;
    ngroups = nelements/N;
  }

  // platform info
  if (print_clinfo) {
    cout << clinfo();
  }

  // initialise for device on platform, with profiling off
  // this creates a context and command queue
  bool profiling = false;
  CLWrapper clw(platform, device, profiling);

  // create some memory objects on the device
  size_t ArraySize = nelements * sizeof(TYPE);
  size_t SumSize   = ngroups * sizeof(TYPE);
  cout << "sizeof(TYPE) = " << sizeof(TYPE) << endl;
  cout << "ArraySize    = " << ArraySize    << endl;
  cl_mem d_in       = clw.dev_malloc(ArraySize);
  cl_mem d_out      = clw.dev_malloc(ArraySize);
  cl_mem d_expected = clw.dev_malloc(ArraySize);
  cl_mem d_sum      = clw.dev_malloc(SumSize);
  cl_mem d_sumout   = clw.dev_malloc(SumSize);
  cl_mem d_error    = clw.dev_malloc(sizeof(unsigned));

  // build kernels
  cl_kernel k1, kinc1, kcheck;
  {
    ostringstream oss;
    oss << "-I. -DOP1 -DNO_INVARIANTS -DN=" << N << " -DINNER=" << kernel;
    cl_program p = clw.compile("meta.cl", oss.str().c_str());
    k1 = clw.create_kernel(p, "voigtlander_op1");
    kinc1 = clw.create_kernel(p, "inc_op1");
    kcheck = clw.create_kernel(p, "check");
  }
  cl_kernel k2, kinc2;
  {
    ostringstream oss;
    oss << "-I. -DOP2 -DNO_INVARIANTS -DN=" << N << " -DINNER=" << kernel;
    cl_program p = clw.compile("meta.cl", oss.str().c_str());
    k2 = clw.create_kernel(p, "voigtlander_op2");
    kinc2 = clw.create_kernel(p, "inc_op2");
  }
  cl_kernel kscan;
  {
    ostringstream oss;
    oss << "-I. -DABSTRACT -DNO_INVARIANTS -DN=" << ngroups;
    cl_program p = clw.compile(kernel.c_str(), oss.str().c_str());
    kscan = clw.create_kernel(p, "prefixsum");
  }

  unsigned long num_op1_vectors = triangular(nelements);
  unsigned long num_op2_vectors = nelements - 1;
  cout << "Will run testvectors "
       << "(op1: " << num_op1_vectors << ") "
       << "(op2: " << num_op2_vectors << ")."
       << endl;

  unsigned error = 0;
  unsigned nvectors = 0;
  vector<TYPE> v(nelements); vector<TYPE> e(nelements);
  for (unsigned place_one=0; place_one<nelements; ++place_one) {
    for (unsigned place_two=place_one+1; place_two<nelements; ++place_two) {
      fill_n(v.begin(), nelements, 0);
      fill_n(e.begin(), nelements, 0);
      v[place_one] = 1;
      fill_n(v.begin() + place_two, nelements-place_two, 2);
      fill_n(e.begin() + place_one, place_two-place_one, 1);
      fill_n(e.begin() + place_two, nelements-place_two, 2);

      error |= runvector(clw,
                kernel,
                k1, kscan, kinc1, kcheck,
                &v[0], &e[0],
                d_in, d_out, d_sum, d_sumout, d_expected, d_error,
                nelements, N, ngroups,
                force_host_scan, print_vectors, print_results, check_random);
      ++nvectors;
    }
    {
      fill_n(v.begin(), nelements, 0);
      fill_n(e.begin(), nelements, 0);
      v[place_one] = 1;
      fill_n(e.begin() + place_one, nelements-place_one, 1);

      error |= runvector(clw,
                kernel,
                k1, kscan, kinc1, kcheck,
                &v[0], &e[0],
                d_in, d_out, d_sum, d_sumout, d_expected, d_error,
                nelements, N, ngroups,
                force_host_scan, print_vectors, print_results, check_random);
      ++nvectors;
    }
  }
  if (error == 0) {
    printf("TEST PASSED (OP1)\n");
  } else {
    printf("TEST FAILED (OP1)\n");
  }
  assert(nvectors == num_op1_vectors);

  error = 0;
  nvectors = 0;
  for (unsigned place_one_two=0; place_one_two<nelements-1; ++place_one_two) {
    fill_n(v.begin(), nelements, 0);
    fill_n(e.begin(), nelements, 0);
    v[place_one_two] = 1;
    v[place_one_two+1] = 2;
    e[place_one_two] = 1;
    fill_n(e.begin() + place_one_two+1, nelements-place_one_two-1, 2);
    error |= runvector(clw,
              kernel,
              k2, kscan, kinc2, kcheck,
              &v[0], &e[0],
              d_in, d_out, d_sum, d_sumout, d_expected, d_error,
              nelements, N, ngroups,
              force_host_scan, print_vectors, print_results, check_random);
    ++nvectors;
  }
  assert(nvectors == num_op2_vectors);
  if (error == 0) {
    printf("TEST PASSED (OP2)\n");
  } else {
    printf("TEST FAILED (OP2)\n");
  }

  return error;
}

